// kate: replace-tabs off; space-indent off;

#include "inotifyutils.h"
#include "../config.h"

#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <assert.h>
#include <errno.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>

#include "inotify.h"
#include "inotify-syscalls.h"
#include "inotify-masks.h"

#define MAX_EVENTS 4096
#define WATCHES_SIZE_PATH "/proc/sys/fs/inotify/max_user_watches"

static int inotify_fd;
static unsigned int * hit_access;
static unsigned int * hit_modify;
static unsigned int * hit_attrib;
static unsigned int * hit_close_nowrite;
static unsigned int * hit_close_write;
static unsigned int * hit_open;
static unsigned int * hit_move_self;
static unsigned int * hit_moved_from;
static unsigned int * hit_moved_to;
static unsigned int * hit_create;
static unsigned int * hit_delete;
static unsigned int * hit_delete_self;
static unsigned int * hit_unmount;
static unsigned int * hit_total;
static unsigned int num_access;
static unsigned int num_modify;
static unsigned int num_attrib;
static unsigned int num_close_nowrite;
static unsigned int num_close_write;
static unsigned int num_open;
static unsigned int num_move_self;
static unsigned int num_moved_to;
static unsigned int num_moved_from;
static unsigned int num_create;
static unsigned int num_delete;
static unsigned int num_delete_self;
static unsigned int num_unmount;
static unsigned int num_total;
static bool collect_stats;
static unsigned int WATCHES_SIZE;
static char ** watches;
static int error;
static unsigned int num_watches;

void inotifyutils_record_stats( struct inotify_event const * event );

char * chrtostr(char ch) {
	static char str[2] = { '\0', '\0' };
	str[0] = ch;
	return str;
}

/**
 * Initialise inotify.
 * @return true on success, false on failure.
 */
bool inotifyutils_initialize() {
	error = 0;
	// Try to initialise inotify
	inotify_fd = inotify_init();
	if (inotify_fd < 0)	{
		fprintf(stderr, "inotify_init() failed.  Are you running Linux 2.6.13 "
		                "or later, and was the\n"
		                "CONFIG_INOTIFY option enabled when your kernel was "
		                "compiled?  If so, \n"
		                "something mysterious has gone wrong.  Please e-mail "
		                PACKAGE_BUGREPORT "\n"
		                " and mention that you saw this message.\n");
		error = inotify_fd;
		return false;
	}

	FILE * watch_limit_file = fopen( WATCHES_SIZE_PATH, "r" );
	if ( !watch_limit_file ) {
		error = errno;
		return false;
	}

	if ( EOF == fscanf( watch_limit_file, "%d", &WATCHES_SIZE ) ) {
		error = errno;
		return false;
	}

	fclose( watch_limit_file );

	watches = (char **)malloc( WATCHES_SIZE*sizeof(char *) );

	collect_stats = false;
	num_watches = 0;

	return true;
}

void inotifyutils_initialize_stats() {
	hit_access = (unsigned int *)calloc( WATCHES_SIZE, sizeof(unsigned int *) );
	hit_modify = (unsigned int *)calloc( WATCHES_SIZE, sizeof(unsigned int *) );
	hit_attrib = (unsigned int *)calloc( WATCHES_SIZE, sizeof(unsigned int *) );
	hit_close_nowrite = (unsigned int *)calloc( WATCHES_SIZE,
	                                            sizeof(unsigned int *) );
	hit_close_write = (unsigned int *)calloc( WATCHES_SIZE,
	                                          sizeof(unsigned int *) );
	hit_open = (unsigned int *)calloc( WATCHES_SIZE,
	                                   sizeof(unsigned int *) );
	hit_move_self = (unsigned int *)calloc( WATCHES_SIZE,
	                                        sizeof(unsigned int *) );
	hit_moved_from = (unsigned int *)calloc( WATCHES_SIZE,
	                                         sizeof(unsigned int *) );
	hit_moved_to = (unsigned int *)calloc( WATCHES_SIZE,
	                                       sizeof(unsigned int *) );
	hit_create = (unsigned int *)calloc( WATCHES_SIZE, sizeof(unsigned int *) );
	hit_delete = (unsigned int *)calloc( WATCHES_SIZE, sizeof(unsigned int *) );
	hit_delete_self = (unsigned int *)calloc( WATCHES_SIZE,
	                                          sizeof(unsigned int *) );
	hit_unmount = (unsigned int *)calloc( WATCHES_SIZE,
	                                      sizeof(unsigned int *) );
	hit_total = (unsigned int *)calloc( WATCHES_SIZE, sizeof(unsigned int *) );

	num_access = 0;
	num_modify = 0;
	num_attrib = 0;
	num_close_nowrite = 0;
	num_close_write = 0;
	num_open = 0;
	num_move_self = 0;
	num_moved_from = 0;
	num_moved_to = 0;
	num_create = 0;
	num_delete = 0;
	num_delete_self = 0;
	num_unmount = 0;
	num_total = 0;

	collect_stats = true;
}

/**
 * Convert event from string form to integer form (as in inotify.h).
 *
 * @param    event    event in string form as defined in inotify.h without
 *                    leading IN_ prefix (e.g., MODIFY, ATTRIB)
 * @return            integer representing the mask specified by 'event', or -1
 *                    if event is invalid
 */
int inotifyutils_str_to_event(char const * event)
{
	static int ret;
	ret = -1;

	if ( 0 == strcasecmp(event, "ACCESS") )
		ret = IN_ACCESS;
	else if ( 0 == strcasecmp(event, "MODIFY") )
		ret = IN_MODIFY;
	else if ( 0 == strcasecmp(event, "ATTRIB") )
		ret = IN_ATTRIB;
	else if ( 0 == strcasecmp(event, "CLOSE_WRITE") )
		ret = IN_CLOSE_WRITE;
	else if ( 0 == strcasecmp(event, "CLOSE_NOWRITE") )
		ret = IN_CLOSE_NOWRITE;
	else if ( 0 == strcasecmp(event, "OPEN") )
		ret = IN_OPEN;
	else if ( 0 == strcasecmp(event, "MOVED_FROM") )
		ret = IN_MOVED_FROM;
	else if ( 0 == strcasecmp(event, "MOVED_TO") )
		ret = IN_MOVED_TO;
	else if ( 0 == strcasecmp(event, "CREATE") )
		ret = IN_CREATE;
	else if ( 0 == strcasecmp(event, "DELETE") )
		ret = IN_DELETE;
	else if ( 0 == strcasecmp(event, "DELETE_SELF") )
		ret = IN_DELETE_SELF;
	else if ( 0 == strcasecmp(event, "UNMOUNT") )
		ret = IN_UNMOUNT;
	else if ( 0 == strcasecmp(event, "Q_OVERFLOW") )
		ret = IN_Q_OVERFLOW;
	else if ( 0 == strcasecmp(event, "IGNORED") )
		ret = IN_IGNORED;
	else if ( 0 == strcasecmp(event, "CLOSE") )
		ret = IN_CLOSE;
	else if ( 0 == strcasecmp(event, "MOVE_SELF") )
		ret = IN_MOVE_SELF;
	else if ( 0 == strcasecmp(event, "MOVE") )
		ret = IN_MOVE;
	else if ( 0 == strcasecmp(event, "ISDIR") )
		ret = IN_ISDIR;
	else if ( 0 == strcasecmp(event, "ONESHOT") )
		ret = IN_ONESHOT;
	else if ( 0 == strcasecmp(event, "ALL_EVENTS") )
		ret = IN_ALL_EVENTS;

	return ret;
}


char * inotifyutils_event_to_str(int events) {
	return inotifyutils_event_to_str_sep(events, ',');
}

/**
 * Convert event from integer form to string form (as in inotify.h).
 * The returned string is from static storage; subsequent calls to this function
 * may overwrite it.
 *
 * @param    events   OR'd event(s) in integer form as defined in inotify.h
 * @param    sep      character used to separate events
 * @return            string representing the event(s)
 */
char * inotifyutils_event_to_str_sep(int events, char sep)
{
	static char ret[1024];
	ret[0] = '\0';
	ret[1] = '\0';

	if ( IN_ACCESS & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "ACCESS" );
	}
	if ( IN_MODIFY & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "MODIFY" );
	}
	if ( IN_ATTRIB & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "ATTRIB" );
	}
	if ( IN_CLOSE_WRITE & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "CLOSE_WRITE" );
	}
	if ( IN_CLOSE_NOWRITE & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "CLOSE_NOWRITE" );
	}
	if ( IN_OPEN & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "OPEN" );
	}
	if ( IN_MOVED_FROM & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "MOVED_FROM" );
	}
	if ( IN_MOVED_TO & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "MOVED_TO" );
	}
	if ( IN_CREATE & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "CREATE" );
	}
	if ( IN_DELETE & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "DELETE" );
	}
	if ( IN_DELETE_SELF & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "DELETE_SELF" );
	}
	if ( IN_UNMOUNT & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "UNMOUNT" );
	}
	if ( IN_Q_OVERFLOW & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "Q_OVERFLOW" );
	}
	if ( IN_IGNORED & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "IGNORED" );
	}
	if ( IN_CLOSE & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "CLOSE" );
	}
	if ( IN_MOVE_SELF & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "MOVE_SELF" );
	}
	if ( IN_ISDIR & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "ISDIR" );
	}
	if ( IN_ONESHOT & events ) {
		strcat( ret, chrtostr(sep) );
		strcat( ret, "ONESHOT" );
	}

	// Maybe we didn't match any... ?
	if ( (ret[0] == '\0') && ( -1 == sprintf( ret, "%c0x%08x", sep,
	      events ) ) ) {
		assert( false );
	}

	return &ret[1];
}

/**
 * Get the filename for a given watch descriptor.
 * @param wd watch descriptor.
 * @return Filename watched by given wd, or NULL if no watch exists with that
 *         wd.
 */
char * inotifyutils_filename_from_wd( int wd ) {
	static char * bad = "(unknown)";
	if ( wd <= 0 ) return bad;
	return watches[wd-1];
}

int inotifyutils_wd_from_filename( char const * filename ) {
	assert( filename );
	static int i;
	for ( i = 0; i < WATCHES_SIZE; ++i ) {
		if ( 0 == strcmp( filename, watches[i] ) ) {
			return i+1;
		}
	}
	return -1;
}

bool inotifyutils_remove_watch_by_filename( char const * filename ) {
	return inotifyutils_remove_watch_by_wd( inotifyutils_wd_from_filename(
	       filename ) );
}

bool inotifyutils_remove_watch_by_wd( int wd ) {
	assert( wd > 0);
	assert( watches[wd - 1] );
	error = 0;
	static int status;
	status = inotify_rm_watch( inotify_fd, wd );
	if ( status < 0 ) {
		fprintf(stderr, "Failed to remove watch on %s: %s\n", watches[wd - 1],
		        strerror(status) );
		error = status;
		return false;
	}
	free( watches[wd - 1] );
	watches[wd - 1] = NULL;
	num_watches--;
	return true;
}

bool inotifyutils_watch_file( char const * filename, int events ) {
	static char const * filenames[2];
	filenames[0] = filename;
	filenames[1] = NULL;
	return inotifyutils_watch_files( filenames, events );
}

bool inotifyutils_watch_files( char const * const * filenames, int events ) {
	assert( filenames );
	error = 0;

	static int i;
	for ( i = 0; filenames[i]; ++i ) {
		static int wd;
		wd = inotify_add_watch( inotify_fd, filenames[i], events );
		if ( wd < 0 ) {
			if ( wd == -1 ) {
				error = errno;
				return false;
			} // if ( wd == -1 )
			else {
				fprintf( stderr, "Failed to watch %s: returned wd was %d "
				         "(expected -1 or >0 )", filenames[i], wd );
				// no appropriate value for error
				return false;
			} // else
		} // if ( wd < 0 )

		// Always end filename with / if it is a directory
		if ( !inotifyutils_isdir(filenames[i])
		     || filenames[i][strlen(filenames[i])-1] == '/') {
			watches[wd - 1] = strdup(filenames[i]);
		}
		else {
			asprintf( &watches[wd - 1], "%s/", filenames[i] );
		}
		num_watches++;
	} // for

	return true;
}

struct inotify_event * inotifyutils_next_event( int timeout ) {

	static struct inotify_event event[MAX_EVENTS];
	static struct inotify_event * ret;
	static int first_byte = 0;
	static ssize_t bytes;

	error = 0;

	if ( first_byte != 0 ) {
		ret = (struct inotify_event *)((char *)&event[0] + first_byte);
		first_byte += sizeof(struct inotify_event) + ret->len;
		if ( first_byte >= bytes ) {
			first_byte = 0;
		}
		if ( collect_stats ) {
			inotifyutils_record_stats( ret );
		}
		return ret;
	}


	static ssize_t this_bytes;
	static int rc;
	static fd_set read_fds;

	static struct timeval read_timeout;
	read_timeout.tv_sec = timeout;
	read_timeout.tv_usec = 0;
	static struct timeval * read_timeout_ptr;
	read_timeout_ptr = ( timeout <= 0 ? NULL : &read_timeout );

	bytes = 0;
	while ( bytes < sizeof(struct inotify_event) ) {
		FD_ZERO(&read_fds);
		FD_SET(inotify_fd, &read_fds);
		rc = select(inotify_fd + 1, &read_fds,
		            NULL, NULL, read_timeout_ptr);
		if ( rc < 0 ) {
			error = errno;
			//perror("Select on inotify failed");
			return NULL;
		}
		else if ( rc == 0 ) {
			// timeout
			return NULL;
		}
		this_bytes = read(inotify_fd, &event[0] + bytes,
		                  sizeof(struct inotify_event)*MAX_EVENTS - bytes);
		if ( this_bytes < 0 ) {
			error = errno;
			//perror("Read from inotify failed");
			return NULL;
		}
		if ( this_bytes == 0 ) {
			fprintf(stderr, "Inotify reported end-of-file.  Possibly too many "
			        "events occurred at once.\n");
			return NULL;
		}
		bytes += this_bytes;
	}

	ret = &event[0];
	first_byte = sizeof(struct inotify_event) + ret->len;
	if ( first_byte >= bytes ) {
		first_byte = 0;
	}
	if ( collect_stats ) {
		inotifyutils_record_stats( ret );
	}
	return ret;
}

unsigned int inotifyutils_watch_recursively( char const * path, int events ) {
	DIR * dir;
	char * my_path;
	error = 0;
	dir = opendir( path );
	if ( !dir ) {
		// If not a directory, don't need to do anything special
		if ( errno == ENOTDIR ) {
			return (inotifyutils_watch_file( path, events )) ? 1 : 0;
		}
		else {
			error = errno;
			return 0;
		}
	}

	if ( path[strlen(path)-1] != '/' ) {
		if ( -1 == asprintf( &my_path, "%s/", path ) ) {
			assert( false );
		}
	}
	else {
		my_path = (char *)path;
	}

	static struct dirent * ent;
	char * next_file;
	static struct stat64 my_stat;
	ent = readdir( dir );
	// Watch each directory within this directory
	while ( ent ) {
		if ( (0 != strcmp( ent->d_name, "." )) &&
		     (0 != strcmp( ent->d_name, ".." )) ) {
			if ( -1 == asprintf(&next_file,"%s%s", my_path, ent->d_name) ) {
				assert( false );
			}
			if ( -1 == lstat64( next_file, &my_stat ) ) {
				error = errno;
				printf( "%s:%d, %s %s\n", __FILE__, __LINE__, next_file,
				        strerror( error ) );
				free( next_file );
				if ( errno != EACCES ) {
					error = errno;
					if ( my_path != path ) free( my_path );
					closedir( dir );
					return 0;
				}
			}
			else if ( S_ISDIR( my_stat.st_mode ) &&
			          !S_ISLNK( my_stat.st_mode )) {
				free( next_file );
				if ( -1 == asprintf(&next_file,"%s%s/",
				     my_path, ent->d_name) ) {
					assert( false );
				}
				static bool status;
				status = inotifyutils_watch_recursively( next_file, events );
				// For some errors, we will continue.
				if ( !status && (EACCES != error) && (ENOENT != error) &&
				     (ELOOP != error) ) {
					printf( "%s:%d, %s %s\n", __FILE__, __LINE__, next_file,
					        strerror( error ) );
					free( next_file );
					if ( my_path != path ) free( my_path );
					closedir( dir );
					return 0;
				}
				free( next_file );
			}
			else {
				free( next_file );
			}
		}
		ent = readdir( dir );
		error = 0;
	}

	if ( my_path != path ) free( my_path );
	closedir( dir );

	return (inotifyutils_watch_file( path, events )) ? num_watches : 0;
}

void inotifyutils_record_stats( struct inotify_event const * event ) {

	if ( IN_ACCESS & event->mask ) {
		hit_access[event->wd - 1]++;
		num_access++;
	}
	if ( IN_MODIFY & event->mask ) {
		hit_modify[event->wd - 1]++;
		num_modify++;
	}
	if ( IN_ATTRIB & event->mask ) {
		hit_attrib[event->wd - 1]++;
		num_attrib++;
	}
	if ( IN_CLOSE_WRITE & event->mask ) {
		hit_close_write[event->wd - 1]++;
		num_close_write++;
	}
	if ( IN_CLOSE_NOWRITE & event->mask ) {
		hit_close_nowrite[event->wd - 1]++;
		num_close_nowrite++;
	}
	if ( IN_OPEN & event->mask ) {
		hit_open[event->wd - 1]++;
		num_open++;
	}
	if ( IN_MOVED_FROM & event->mask ) {
		hit_moved_from[event->wd - 1]++;;
		num_moved_from++;
	}
	if ( IN_MOVED_TO & event->mask ) {
		hit_moved_to[event->wd - 1]++;
		num_moved_to++;
	}
	if ( IN_CREATE & event->mask ) {
		hit_create[event->wd - 1]++;
		num_create++;
	}
	if ( IN_DELETE & event->mask ) {
		hit_delete[event->wd - 1]++;
		num_delete++;
	}
	if ( IN_DELETE_SELF & event->mask ) {
		hit_delete_self[event->wd - 1]++;
		num_delete_self++;
	}
	if ( IN_UNMOUNT & event->mask ) {
		hit_unmount[event->wd - 1]++;
		num_unmount++;
	}
	if ( IN_MOVE_SELF & event->mask ) {
		hit_move_self[event->wd - 1]++;
		num_move_self++;
	}

	hit_total[event->wd - 1]++;
	num_total++;

}

unsigned int inotifyutils_get_stat_by_wd( int wd, int event ) {
	if ( IN_ACCESS == event )
		return hit_access[wd - 1];
	if ( IN_MODIFY == event )
		return hit_modify[wd - 1];
	if ( IN_ATTRIB == event )
		return hit_attrib[wd - 1];
	if ( IN_CLOSE_WRITE == event )
		return hit_close_write[wd - 1];
	if ( IN_CLOSE_NOWRITE == event )
		return hit_close_nowrite[wd - 1];
	if ( IN_OPEN == event )
		return hit_open[wd - 1];
	if ( IN_MOVED_FROM == event )
		return hit_moved_from[wd - 1];;
	if ( IN_MOVED_TO == event )
		return hit_moved_to[wd - 1];
	if ( IN_CREATE == event )
		return hit_create[wd - 1];
	if ( IN_DELETE == event )
		return hit_delete[wd - 1];
	if ( IN_DELETE_SELF == event )
		return hit_delete_self[wd - 1];
	if ( IN_UNMOUNT == event )
		return hit_unmount[wd - 1];
	if ( IN_MOVE_SELF == event )
		return hit_move_self[wd - 1];

	return hit_total[wd - 1];
}

unsigned int inotifyutils_get_stat_total( int event ) {
	if ( IN_ACCESS == event )
		return num_access;
	if ( IN_MODIFY == event )
		return num_modify;
	if ( IN_ATTRIB == event )
		return num_attrib;
	if ( IN_CLOSE_WRITE == event )
		return num_close_write;
	if ( IN_CLOSE_NOWRITE == event )
		return num_close_nowrite;
	if ( IN_OPEN == event )
		return num_open;
	if ( IN_MOVED_FROM == event )
		return num_moved_from;
	if ( IN_MOVED_TO == event )
		return num_moved_to;
	if ( IN_CREATE == event )
		return num_create;
	if ( IN_DELETE == event )
		return num_delete;
	if ( IN_DELETE_SELF == event )
		return num_delete_self;
	if ( IN_UNMOUNT == event )
		return num_unmount;
	if ( IN_MOVE_SELF == event )
		return num_move_self;

	return num_total;
}

unsigned int inotifyutils_get_stat_by_filename( char const * filename,
                                                int event ) {
	return inotifyutils_get_stat_by_wd( inotifyutils_wd_from_filename(
	       filename ), event );
}

int inotifyutils_error() {
	return error;
}

void inotifyutils_print_event_descriptions() {
	printf("\taccess\t\tfile or directory contents were read\n");
	printf("\tmodify\t\tfile or directory contents were written\n");
	printf("\tattrib\t\tfile or directory attributes changed\n");
	printf("\tclose_write\tfile or directory closed, after being opened in\n"
	       "\t           \twriteable mode\n");
	printf("\tclose_nowrite\tfile or directory closed, after being opened in\n"
	       "\t           \tread-only mode\n");
	printf("\tclose\t\tfile or directory closed, regardless of read/write "
	       "mode\n");
	printf("\topen\t\tfile or directory opened\n");
	printf("\tmoved_to\tfile or directory moved to watched directory\n");
	printf("\tmoved_from\tfile or directory moved from watched directory\n");
	printf("\tmove\t\tfile or directory moved to or from watched directory\n");
	printf("\tcreate\t\tfile or directory created within watched directory\n");
	printf("\tdelete\t\tfile or directory deleted within watched directory\n");
	printf("\tdelete_self\tfile or directory was deleted\n");
	printf("\tunmount\t\tfile system containing file or directory unmounted\n");
}

bool inotifyutils_isdir( char const * path ) {
	static struct stat64 my_stat;

	if ( -1 == lstat64( path, &my_stat ) ) {
		fprintf(stderr, "Stat failed on %s: %s\n", path, strerror(errno));
		return 0;
	}

	return S_ISDIR( my_stat.st_mode ) && !S_ISLNK( my_stat.st_mode );
}
